body {
  font-family: "Roboto", sans-serif;
  line-height: 1.6;
  color: #333;
  max-width: 1200px;
  margin: 0 auto;
  padding: 20px;
  background-color: #f5f5f5;
}

h1,
h2 {
  color: #2c3e50;
  margin-bottom: 20px;
}

.container {
  display: flex;
  justify-content: space-between;
  margin-top: 30px;
}

.column {
  background: white;
  border-radius: 8px;
  padding: 25px;
  box-shadow: 0 4px 6px rgba(0, 0, 0, 0.1);
  width: 48%;
}

button {
  background-color: #3498db;
  color: white;
  border: none;
  padding: 10px 15px;
  margin: 10px 5px;
  border-radius: 5px;
  cursor: pointer;
  transition: background-color 0.3s;
}

button:hover {
  background-color: #2980b9;
}

input {
  padding: 10px;
  margin: 10px 5px;
  border: 1px solid #ddd;
  border-radius: 4px;
  width: calc(100% - 130px);
}

.plugin-item,
.running-item {
  background-color: #ecf0f1;
  margin: 15px 0;
  padding: 20px;
  border-radius: 5px;
  display: flex;
  flex-direction: column;
}

.plugin-item h3 {
  margin-top: 0;
  margin-bottom: 10px;
}

.plugin-description {
  margin-bottom: 15px;
}

.plugin-controls {
  display: flex;
  justify-content: space-between;
  align-items: center;
}

.state-viewer {
  margin-top: 30px;
  background: white;
  border-radius: 8px;
  padding: 25px;
  box-shadow: 0 4px 6px rgba(0, 0, 0, 0.1);
}

.state-viewer > div {
  margin-bottom: 15px;
}

.state-viewer label {
  display: block;
  margin-bottom: 5px;
}

#stateDisplay {
  white-space: pre-wrap;
  background-color: #ecf0f1;
  padding: 20px;
  border-radius: 5px;
  margin-top: 20px;
}

.datasite-manager {
  margin-top: 30px;
  background: white;
  border-radius: 8px;
  padding: 25px;
  box-shadow: 0 4px 6px rgba(0, 0, 0, 0.1);
}

.datasite-manager input[type="text"] {
  width: calc(100% - 120px);
  margin-right: 10px;
}

#datasiteList {
  list-style-type: none;
  padding: 0;
}

#datasiteList li {
  display: flex;
  justify-content: space-between;
  align-items: center;
  padding: 10px 0;
  border-bottom: 1px solid #eee;
}

#datasiteList li:last-child {
  border-bottom: none;
}

#datasiteList button {
  background-color: #e74c3c;
}

#datasiteList button:hover {
  background-color: #c0392b;
}
